import logging
import pathlib

import numpy as np
import torch
from torch import Tensor

import sd_mecha.builtin_model_archs
import sd_mecha.builtin_model_types
from typing import Optional, Dict, Mapping
from sd_mecha.recipe_merger import RecipeMerger
from sd_mecha import recipe_nodes, merge_methods, extensions
from sd_mecha.extensions.merge_method import RecipeNodeOrPath, path_to_node
from sd_mecha.recipe_nodes import MergeSpace
from sd_mecha.hypers import Hyper, classes, blocks, default
from sd_mecha.recipe_serializer import serialize, deserialize, deserialize_path


def merge_and_save(
    recipe: recipe_nodes.RecipeNode,
    models_dir: pathlib.Path,
    output_path: pathlib.Path,
):
    merger = RecipeMerger(models_dir=models_dir)
    merger.merge_and_save(recipe, output=output_path)


def serialize_and_save(
    recipe: recipe_nodes.RecipeNode,
    output_path: pathlib.Path | str,
):
    serialized = serialize(recipe)

    if not isinstance(output_path, pathlib.Path):
        output_path = pathlib.Path(output_path)
    if not output_path.suffix:
        output_path = output_path.with_suffix(".mecha")
    output_path = output_path.absolute()

    logging.info(f"Saving recipe to {output_path}")
    with open(output_path, "w") as f:
        f.write(serialized)


weighted_sum = merge_methods.weighted_sum
slerp = merge_methods.slerp


def add_difference(
    a: RecipeNodeOrPath, b: RecipeNodeOrPath, c: Optional[RecipeNodeOrPath] = None, *,
    alpha: Hyper = 1.0,
    clip_to_ab: Optional[bool] = None,
    device: Optional[str] = None,
    dtype: Optional[torch.dtype] = None,
) -> recipe_nodes.RecipeNode:
    a: recipe_nodes.RecipeNode = extensions.merge_method.path_to_node(a)
    b: recipe_nodes.RecipeNode = extensions.merge_method.path_to_node(b)
    original_b = b

    if c is not None:
        c = extensions.merge_method.path_to_node(c)
        b = subtract(
            b, c,
            device=device,
            dtype=dtype,
        )

    res = merge_methods.add_difference(
        a, b,
        alpha=alpha,
        device=device,
        dtype=dtype,
    )

    if a.merge_space == original_b.merge_space:
        b = original_b

    if clip_to_ab is None:
        clip_to_ab = a.merge_space == b.merge_space

    if clip_to_ab:
        if a.merge_space != b.merge_space:
            raise TypeError(f"Merge space of A {a.merge_space} and B {b.merge_space} must be the same to clip the merge.")
        res = clip(
            res, a, b,
            device=device,
            dtype=dtype,
        )

    return res


subtract = merge_methods.subtract
perpendicular_component = merge_methods.perpendicular_component


def add_perpendicular(
    a: RecipeNodeOrPath, b: RecipeNodeOrPath, c: RecipeNodeOrPath, *,
    alpha: Hyper = 1.0,
    device: Optional[str] = None,
    dtype: Optional[torch.dtype] = None,
) -> recipe_nodes.RecipeNode:
    a = path_to_node(a)
    b = path_to_node(b)
    c = path_to_node(c)

    a_diff = subtract(
        a, c,
        device=device,
        dtype=dtype,
    )
    b_diff = subtract(
        b, c,
        device=device,
        dtype=dtype,
    )

    perp_diff = merge_methods.perpendicular_component(
        a_diff, b_diff,
        device=device,
        dtype=dtype,
    )

    return merge_methods.add_difference(
        a, perp_diff,
        alpha=alpha,
        device=device,
        dtype=dtype,
    )


geometric_sum = merge_methods.geometric_sum
train_difference = merge_methods.train_difference
cosine_add_a = merge_methods.add_cosine_a
cosine_add_b = merge_methods.add_cosine_b
ties_sum = merge_methods.ties_sum


# latex notes in reference to original implementation: https://arxiv.org/abs/2306.01708
# - `base`: $$ \theta_{init} $$
# - `*models`: $$ \{\theta_{init}\}_{t=1}^n $$
# - `models` after `subtract`: $$ \tau_t $$
# - `alpha`: $$ \lambda $$
# - `k`: $$ k $$ ( From $$ \% $$ to $$ 1 $$ )
# - `res`: $$ \lambda * \tau_m $$
# - `return`: $$ \theta_m $$
def add_difference_ties(
    base: RecipeNodeOrPath,
    *models: RecipeNodeOrPath,
    alpha: float,
    k: float = 0.2,
    device: Optional[str] = None,
    dtype: Optional[torch.dtype] = None,
) -> recipe_nodes.RecipeNode:
    # $$ \{\theta_{init}\}_{t=1}^n $$
    base = path_to_node(base)
    models = tuple(path_to_node(model) for model in models)

    # Create task vectors.
    # $$ \tau_t $$
    models = tuple(
        subtract(model, base)
        if model.merge_space is MergeSpace.BASE else
        model
        for model in models
    )

    # step 1 + step 2 + step 3
    res = ties_sum(
        *models,
        k=k,
        device=device,
        dtype=dtype,
    )

    # Obtain merged checkpoint

    # $$ \theta_{init} + \lambda * \tau_m $$
    return add_difference(
        base, res,
        alpha=alpha,
        device=device,
        dtype=dtype,
    )


def copy_region(
    a: RecipeNodeOrPath, b: RecipeNodeOrPath, c: Optional[RecipeNodeOrPath] = None, *,
    width: Hyper = 1.0,
    offset: Hyper = 0.0,
    top_k: bool = False,
    device: Optional[str] = "cuda",
    dtype: Optional[torch.dtype] = None,
) -> recipe_nodes.RecipeNode:
    a = path_to_node(a)
    b = path_to_node(b)

    if c is not None:
        c = path_to_node(c)

        a = subtract(
            a, c,
            device=device,
            dtype=dtype,
        )
        b = subtract(
            b, c,
            device=device,
            dtype=dtype,
        )

    copy_method = [merge_methods.tensor_sum, merge_methods.top_k_tensor_sum][int(top_k)]
    res = copy_method(
        a, b,
        width=width,
        offset=offset,
        device=device,
        dtype=dtype,
    )

    if c is not None:
        res = merge_methods.add_difference(
            c, res,
            alpha=1.0,
            device=device,
            dtype=dtype,
        )

    return res


tensor_sum = copy_region
distribution_crossover = merge_methods.distribution_crossover
crossover = merge_methods.crossover


def rotate(
    a: RecipeNodeOrPath, b: RecipeNodeOrPath, c: Optional[RecipeNodeOrPath] = None, *,
    alpha: Hyper = 1.0,
    beta: Hyper = 0.0,
    device: Optional[str] = "cuda",
    dtype: Optional[torch.dtype] = None,
    cache: Optional[Dict[str, torch.Tensor]] = None,
) -> recipe_nodes.RecipeNode:
    a = path_to_node(a)
    b = path_to_node(b)

    if c is not None:
        c = path_to_node(c)

        a = subtract(
            a, c,
            device=device,
            dtype=dtype,
        )
        b = subtract(
            b, c,
            device=device,
            dtype=dtype,
        )

    res = merge_methods.rotate(
        a, b,
        alignment=alpha,
        alpha=beta,
        cache=cache,
        device=device,
        dtype=dtype,
    )

    if c is not None:
        res = merge_methods.add_difference(
            c, res,
            alpha=1.0,
            device=device,
            dtype=dtype,
        )

    return res


clip = merge_methods.clip


def dropout(
    a: RecipeNodeOrPath,
    *models: RecipeNodeOrPath,
    probability: Hyper = 0.9,
    alpha: Hyper = 0.5,
    overlap: Hyper = 0.0,
    overlap_emphasis: Hyper = 0.0,
    seed: Optional[Hyper] = None,
    device: Optional[str] = None,
    dtype: Optional[torch.dtype] = None,
) -> recipe_nodes.RecipeNode:
    deltas = [
        subtract(model, a)
        for model in models
    ]
    ba_delta = merge_methods.dropout(*deltas, probability=probability, overlap=overlap, overlap_skew=overlap_emphasis, seed=seed, device=device, dtype=dtype)
    return sd_mecha.add_difference(a, ba_delta, alpha=alpha, device=device, dtype=dtype)


def model(state_dict: str | pathlib.Path | Mapping[str, Tensor], model_arch: str = "sd1", model_type: str = "base"):
    return recipe_nodes.ModelRecipeNode(state_dict, model_arch, model_type)


def lora(state_dict: str | pathlib.Path, model_arch: str = "sd1"):
    return recipe_nodes.ModelRecipeNode(state_dict, model_arch, "lora")


def parameter(name: str, merge_space: MergeSpace, model_arch: Optional[str] = None):
    return recipe_nodes.ParameterRecipeNode(name, merge_space, model_arch)


def set_log_level(level: str = "INFO"):
    logging.basicConfig(format="%(levelname)s: %(message)s", level=level)



import functools
import math
import operator
import numpy as np
import torch
from scipy.stats import binom
from torch import Tensor
from typing import Tuple, TypeVar, Dict, Optional
from sd_mecha.hypers import Hyper
from sd_mecha.merge_space import MergeSpace
from sd_mecha.extensions.merge_method import LiftFlag, convert_to_recipe


EPSILON = 1e-10
SameMergeSpace = TypeVar("SameMergeSpace", bound=LiftFlag[MergeSpace.BASE | MergeSpace.DELTA])


@convert_to_recipe
def weighted_sum(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    *,
    alpha: Hyper = 0.5,
    **kwargs,
) -> Tensor | SameMergeSpace:
    return (1 - alpha) * a + alpha * b


@convert_to_recipe
def slerp(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    *,
    alpha: Hyper = 0.5,
    **kwargs,
) -> Tensor | SameMergeSpace:
    a_normalized = a / a.norm()
    b_normalized = b / b.norm()

    ab_dot = (a_normalized * b_normalized).sum().clip(-1, 1)

    if 1 - torch.abs(ab_dot) < EPSILON:
        return weighted_sum.__wrapped__(a, b, alpha=alpha)

    omega = torch.arccos(ab_dot)
    a_contrib = a_normalized * torch.sin((1-alpha)*omega)
    b_contrib = b_normalized * torch.sin(alpha*omega)
    res = (a_contrib + b_contrib) / torch.sin(omega)
    return res * weighted_sum.__wrapped__(a.norm(), b.norm(), alpha=alpha)


@convert_to_recipe
def add_difference(
    a: Tensor | SameMergeSpace,
    b: Tensor | LiftFlag[MergeSpace.DELTA],
    *,
    alpha: Hyper = 0.5,
    **kwargs,
) -> Tensor | SameMergeSpace:
    return a + alpha * b


@convert_to_recipe
def subtract(
    a: Tensor | LiftFlag[MergeSpace.BASE],
    b: Tensor | LiftFlag[MergeSpace.BASE],
    **kwargs,
) -> Tensor | LiftFlag[MergeSpace.DELTA]:
    return a - b


@convert_to_recipe
def perpendicular_component(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    **kwargs,
) -> Tensor | SameMergeSpace:
    norm_a = torch.linalg.norm(a)
    res = b - a * (a / norm_a * (b / norm_a)).sum()
    if res.isnan().any():
        return torch.zeros_like(a)
    return res


@convert_to_recipe
def geometric_sum(
    a: Tensor | LiftFlag[MergeSpace.DELTA],
    b: Tensor | LiftFlag[MergeSpace.DELTA],
    *,
    alpha: Hyper = 0.5,
    **kwargs,
) -> Tensor | LiftFlag[MergeSpace.DELTA]:
    a = torch.complex(a, torch.zeros_like(a))
    b = torch.complex(b, torch.zeros_like(b))
    res = a ** (1 - alpha) * b ** alpha
    return res.real


@convert_to_recipe
def add_cosine_a(
    a: Tensor | LiftFlag[MergeSpace.BASE],
    b: Tensor | LiftFlag[MergeSpace.BASE],
    *,
    alpha: Hyper,
    **kwargs,
) -> Tensor | LiftFlag[MergeSpace.BASE]:
    a_norm = torch.nn.functional.normalize(a, dim=0)
    b_norm = torch.nn.functional.normalize(b, dim=0)
    similarity = torch.nn.functional.cosine_similarity(a_norm, b_norm, dim=0)
    return add_cosine_generic(a, b, alpha, similarity)


@convert_to_recipe
def add_cosine_b(
    a: Tensor | LiftFlag[MergeSpace.BASE],
    b: Tensor | LiftFlag[MergeSpace.BASE],
    *,
    alpha: Hyper,
    **kwargs,
) -> Tensor | LiftFlag[MergeSpace.BASE]:
    similarity = torch.nn.functional.cosine_similarity(a, b, dim=0)
    dot_product = torch.sum(a * b)
    magnitude_similarity = dot_product / (torch.norm(a) * torch.norm(b))
    combined_similarity = (similarity + magnitude_similarity) / 2.0
    return add_cosine_generic(a, b, alpha, combined_similarity)


def add_cosine_generic(a: Tensor, b: Tensor, alpha: float, similarity: Tensor) -> Tensor:
    k = 1 - torch.clamp(similarity - alpha, 0, 1)
    return weighted_sum.__wrapped__(a, b, alpha=k)


# latex notes in reference to original implementation: https://arxiv.org/abs/2306.01708
# - `delta`: $$ \hat{\tau}_t $$
# - `signs`: $$ \gamma_t $$
# - `final_sign`: $$ \gamma_m^p = sgn(\sum_{t=1}^n \hat{\tau}_t^p) $$
# - `delta_filters`: $$ \{ \gamma_t^p = \gamma_m^p \} $$
# - `param_counts`: $$ |A^p| $$
# - `filtered_delta`: $$ \sum_{t\in{A^p}} \hat{\tau}_t^p $$
# - `return`: $$ \lambda * \tau_m $$
@convert_to_recipe
def ties_sum(  # aka add_difference_ties
    *models: Tensor | LiftFlag[MergeSpace.DELTA],
    k: Hyper = 0.2,
    **kwargs,
) -> Tensor | LiftFlag[MergeSpace.DELTA]:

    # Step 1: Trim redundant parameters

    # $$ \hat{\tau}_t $$ O(N) in space
    deltas = [
        # $$ keep_topk_reset_rest_to_zero(\tau_t, k) $$
        filter_top_k(m, k)
        for m in models
    ]
    deltas = torch.stack(deltas, dim=0)

    # Step 2: Elect Final Signs.

    # $$ \gamma_t $$ 
    signs = torch.sign(deltas)

    # $$ \gamma_m^p = sgn(\sum_{t=1}^n \hat{\tau}_t^p) $$
    final_sign = torch.sign(torch.sum(deltas, dim=0)) 

    # Step 3: Disjoint merge.

    # $$ \{ \gamma_t^p = \gamma_m^p \} $$
    delta_filters = (signs == final_sign).float()

    # $$ |A^p| $$
    param_counts = torch.sum(delta_filters, dim=0)

    # $$ \sum_{t\in{A^P}} \hat{\tau}_t^p $$
    filtered_delta = (deltas * delta_filters).sum(dim=0)

    # $$ \tau_m $$
    return torch.nan_to_num(filtered_delta / param_counts)


def filter_top_k(a: Tensor, k: float):
    k = max(int((1 - k) * torch.numel(a)), 1)
    k_value, _ = torch.kthvalue(torch.abs(a.flatten()).float(), k)
    top_k_filter = (torch.abs(a) >= k_value).float()
    return a * top_k_filter


@convert_to_recipe
def tensor_sum(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    *,
    width: Hyper = 0.5,
    offset: Hyper = 0.0,
    **kwargs,
) -> Tensor | SameMergeSpace:
    if a.shape == ():
        if width > 0.5:
            return b
        return a

    start_i, end_i, region_is_inverted = ratio_to_region(width, offset, a.size(0))
    if region_is_inverted:
        b[start_i:end_i] = a[start_i:end_i]
        return b
    else:
        a[start_i:end_i] = b[start_i:end_i]
        return a


@convert_to_recipe
def top_k_tensor_sum(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    *,
    width: Hyper = 0.5,
    offset: Hyper = 0.0,
    **kwargs,
) -> Tensor | SameMergeSpace:
    a_flat = torch.flatten(a)
    a_dist = torch.msort(a_flat)
    b_indices = torch.argsort(torch.flatten(b), stable=True)
    redist_indices = torch.argsort(b_indices)

    start_i, end_i, region_is_inverted = ratio_to_region(width, offset, torch.numel(a))
    start_top_k = kth_abs_value(a_dist, start_i)
    end_top_k = kth_abs_value(a_dist, end_i)

    indices_mask = (start_top_k <= torch.abs(a_dist)) & (torch.abs(a_dist) <= end_top_k)
    if region_is_inverted:
        indices_mask = ~indices_mask
    indices_mask = torch.gather(indices_mask.float(), 0, redist_indices)

    a_redist = torch.gather(a_dist, 0, redist_indices)
    a_redist = (1 - indices_mask) * a_flat + indices_mask * a_redist
    return a_redist.reshape_as(a)


def kth_abs_value(a: Tensor, k: int) -> Tensor:
    if k <= 0:
        return torch.tensor(-1, device=a.device)
    else:
        return torch.kthvalue(torch.abs(a.float()), k)[0]


def ratio_to_region(width: float, offset: float, n: int) -> Tuple[int, int, bool]:
    if width < 0:
        offset += width
        width = -width
    width = min(width, 1)

    if offset < 0:
        offset = 1 + offset - int(offset)
    offset = math.fmod(offset, 1.0)

    if width + offset <= 1:
        inverted = False
        start = offset * n
        end = (width + offset) * n
    else:
        inverted = True
        start = (width + offset - 1) * n
        end = offset * n

    return round(start), round(end), inverted


@convert_to_recipe
def train_difference(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    c: Tensor | SameMergeSpace,
    *,
    alpha: Hyper = 1.0,
    **kwargs,
) -> Tensor | SameMergeSpace:
    threshold = torch.maximum(torch.abs(a - c), torch.abs(b - c))
    dissimilarity = torch.clamp(torch.nan_to_num((c - a) * (b - c) / threshold**2, nan=0), 0)

    return a + (b - c) * alpha * dissimilarity


@convert_to_recipe
def multiply_quotient(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    c: Tensor | SameMergeSpace,
    *,
    alpha: Hyper = 1.0,
    **kwargs,
) -> Tensor | SameMergeSpace:
    ac_log = torch.log(a.abs()) - torch.log(c.abs())
    bc_log = torch.log(b.abs()) - torch.log(c.abs())

    b = torch.complex(b, torch.zeros_like(b))
    c = torch.complex(c, torch.zeros_like(c))

    threshold = torch.maximum(torch.abs(ac_log), torch.abs(bc_log))
    alpha *= torch.clamp(-torch.nan_to_num(ac_log * bc_log / threshold**2, nan=0), 0)

    res = a * (b / c)**alpha
    res = torch.where(torch.isnan(res), a, res)
    del a, b, c
    return res.real


@convert_to_recipe
def distribution_crossover(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    c: Tensor | SameMergeSpace,
    *,
    mean: Hyper,
    tilt: Hyper,
    **kwargs,
) -> Tensor | SameMergeSpace:
    if mean == 0:
        return a
    if tilt == 1 or a.shape == ():
        return weighted_sum.__wrapped__(a, b, alpha=mean)

    c_indices = torch.argsort(torch.flatten(c))
    a_dist = torch.gather(torch.flatten(a), 0, c_indices)
    b_dist = torch.gather(torch.flatten(b), 0, c_indices)

    a_dft = torch.fft.rfft(a_dist)
    b_dft = torch.fft.rfft(b_dist)

    dft_filter = create_filter((a_dft.numel(),), mean, tilt, device=a.device)

    x_dft = (1 - dft_filter) * a_dft + dft_filter * b_dft
    x_dist = torch.fft.irfft(x_dft, a_dist.shape[0])
    x_values = torch.gather(x_dist, 0, torch.argsort(c_indices))
    return x_values.reshape_as(a)


@convert_to_recipe
def crossover(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    *,
    mean: Hyper = 0.5,
    tilt: Hyper = 0.0,
    **kwargs,
) -> Tensor | SameMergeSpace:
    if mean == 0:
        return a
    if tilt == 1:
        return weighted_sum.__wrapped__(a, b, alpha=mean)

    if len(a.shape) == 0 or torch.allclose(a.half(), b.half()):
        return weighted_sum.__wrapped__(a, b, alpha=tilt)

    if a.shape[0] > 40000 or len(a.shape) == 4 and sum(a.shape[2:]) > 2:
        shape = a.shape[1:]
    else:
        shape = a.shape

    a_dft = torch.fft.rfftn(a, s=shape)
    b_dft = torch.fft.rfftn(b, s=shape)

    dft_filter = create_filter(a_dft.shape, mean, tilt, device=a.device)

    x_dft = (1 - dft_filter)*a_dft + dft_filter*b_dft
    return torch.fft.irfftn(x_dft, s=shape)


def create_filter(shape: Tuple[int, ...] | torch.Size, mean: float, tilt: float, steps=100, precision=EPSILON, device=None):
    """
    Create a crossover filter. The cut is first tilted, then slid along its normal to match the mean.
    :param shape: shape of the filter
    :param mean: the mean of the filter. must be in [0, 1]
    :param tilt: tilt of the filter. 0 = vertical filter, 0.5 = 45 degrees, 1 = degenerates to a weighted sum with alpha=mean
    :param steps: maximum number of optimization steps to apply over the mean until the filter converges
    :param precision: the accepted loss between the requested mean and the effective mean of the filter
    :param device: device of the filter
    :return:
    """
    if not 0 <= mean <= 1:
        raise ValueError("filter mean must be between 0 and 1")

    gradients = [
        torch.linspace(0, 1, s, device=device)**2
        for s in shape
    ]

    if len(shape) > 1:
        grids = torch.meshgrid(*gradients, indexing='ij')
        mesh = torch.sqrt(torch.sum(torch.stack(grids), dim=0)) / math.sqrt(len(shape))
    else:
        mesh = gradients[0]

    # train the offset of the cut to pick the right ratio of parameters
    trained_offset = mean
    dft_filter = mesh
    for step in range(steps):
        if tilt < EPSILON:
            dft_filter = (mesh > 1 - trained_offset).float()
        else:
            tilt_cot = 1 / math.tan(math.pi * tilt / 2)
            dft_filter = torch.clamp(mesh*tilt_cot + trained_offset*tilt_cot + trained_offset - tilt_cot, 0, 1)
        current_mean = dft_filter.mean()
        loss = mean - current_mean
        if abs(loss) < precision:
            break
        trained_offset += loss

    return dft_filter


@convert_to_recipe(volatile_hypers=["cache"])
def rotate(
    a: Tensor | SameMergeSpace,
    b: Tensor | SameMergeSpace,
    *,
    alignment: Hyper = 1.0,
    alpha: Hyper = 0.0,
    cache: Optional[Dict[str, Dict[str, Tensor]]] = None,
    **kwargs,
) -> Tensor | SameMergeSpace:
    if alignment == 0 and alpha == 0:
        return a

    if len(a.shape) < 2 or torch.allclose(a.half(), b.half()):
        return weighted_sum.__wrapped__(a, b, alpha=alpha)

    is_conv = len(a.shape) == 4 and a.shape[-1] != 1
    if is_conv:
        shape_2d = (-1, functools.reduce(operator.mul, a.shape[2:]))
    elif len(a.shape) == 4:
        shape_2d = (-1, functools.reduce(operator.mul, a.shape[1:]))
    else:
        shape_2d = (-1, a.shape[-1])

    a_neurons = a.reshape(*shape_2d)
    b_neurons = b.reshape(*shape_2d)
    a_centroid = a_neurons.mean(0)
    b_centroid = b_neurons.mean(0)
    a_neurons -= a_centroid
    b_neurons -= b_centroid

    alignment_is_float = alignment != round(alignment)

    if cache is not None:
        key = kwargs["key"]
        if key not in cache:
            cache[key] = {}
        cache = cache[key]

    if cache is not None and "rotation" in cache:
        rotation = transform = cache["rotation"].to(a.device, a.dtype)
    else:
        rotation = transform = orthogonal_procrustes(a_neurons, b_neurons, cancel_reflection=alignment_is_float)
        if cache is not None:
            cache["rotation"] = rotation.to("cpu", torch.float16)

    if alignment_is_float:
        transform = fractional_matrix_power(transform, alignment, cache)
    elif alignment == 0:
        transform = torch.eye(
            len(transform),
            dtype=transform.dtype,
            device=transform.device,
        )
    elif alignment != 1:
        transform = torch.linalg.matrix_power(transform, round(alignment))

    if alpha != 0:
        # interpolate the relationship between the neurons
        a_neurons = weighted_sum.__wrapped__(a_neurons, b_neurons @ rotation.T, alpha=alpha)

    a_neurons @= transform
    a_neurons += weighted_sum.__wrapped__(a_centroid, b_centroid, alpha=alignment)
    return a_neurons.reshape_as(a)


def orthogonal_procrustes(a, b, cancel_reflection: bool = False):
    svd_driver = "gesvd" if a.is_cuda else None
    u, _, v_t = torch.linalg.svd(a.T @ b, driver=svd_driver)

    if cancel_reflection:
        u[:, -1] /= torch.det(u) * torch.det(v_t)

    transform = u @ v_t
    if not torch.isfinite(u).all():
        raise ValueError(
            f"determinant error: {torch.det(transform)}. "
            'This can happen when merging on the CPU with the "rotate" method. '
            "Consider merging on a cuda device, "
            "or try setting alpha to 1 for the problematic blocks. "
            "See this related discussion for more info: "
            "https://github.com/s1dlx/meh/pull/50#discussion_r1429469484"
        )

    return transform


def fractional_matrix_power(matrix: Tensor, power: float, cache: Optional[Dict[str, Tensor]] = None):
    if cache is not None and "eigenvalues" in cache:
        complex_dtype = torch_complex_dtype_map[matrix.dtype]
        eigenvalues = cache["eigenvalues"].to(matrix.device, complex_dtype)
        eigenvectors = cache["eigenvectors"].to(matrix.device, complex_dtype)
        eigenvectors_inv = cache["eigenvectors_inv"].to(matrix.device, complex_dtype)
    else:
        eigenvalues, eigenvectors = torch.linalg.eig(matrix)
        eigenvectors_inv = torch.linalg.inv(eigenvectors)
        if cache is not None:
            cache["eigenvalues"] = eigenvalues.to("cpu", torch.complex32)
            cache["eigenvectors"] = eigenvectors.to("cpu", torch.complex32)
            cache["eigenvectors_inv"] = eigenvectors_inv.to("cpu", torch.complex32)

    eigenvalues.pow_(power)
    result = eigenvectors @ torch.diag(eigenvalues) @ eigenvectors_inv
    return result.real.to(dtype=matrix.dtype)


torch_complex_dtype_map = {
    torch.bfloat16: torch.complex32,
    torch.float16: torch.complex32,
    torch.float32: torch.complex64,
    torch.float64: torch.complex128,
}


@convert_to_recipe
def clip(
    a: Tensor | SameMergeSpace,
    *bounds: Tensor | SameMergeSpace,
    stiffness: Hyper = 0.0,
    **kwargs,
) -> Tensor | SameMergeSpace:
    maximums = functools.reduce(torch.maximum, bounds)
    minimums = functools.reduce(torch.minimum, bounds)
    centers = (maximums + minimums) / 2

    if stiffness:
        smallest_positive = maximums
        largest_negative = minimums

        for i, bound in enumerate(bounds):
            smallest_positive = torch.where((smallest_positive >= bound) & (bound >= centers), bound, smallest_positive)
            largest_negative = torch.where((largest_negative <= bound) & (bound <= centers), bound, largest_negative)

        maximums = weighted_sum.__wrapped__(maximums, smallest_positive, alpha=stiffness)
        minimums = weighted_sum.__wrapped__(minimums, largest_negative, alpha=stiffness)

    return torch.minimum(torch.maximum(a, minimums), maximums)


@convert_to_recipe
def dropout(  # aka n-supermario
    delta0: Tensor | LiftFlag[MergeSpace.DELTA],
    *deltas: Tensor | LiftFlag[MergeSpace.DELTA],
    probability: Hyper = 0.9,
    overlap: Hyper = 0.0,
    overlap_emphasis: Hyper = 0.0,
    seed: Hyper = None,
    **kwargs,
) -> Tensor | LiftFlag[MergeSpace.DELTA]:
    rng = np.random.default_rng(seed)
    deltas = (delta0,) + deltas

    ks = np.arange(0, 2 ** len(deltas))
    pmf = overlapping_sets_pmf(len(deltas), probability, overlap, overlap_emphasis)
    masks = torch.from_numpy(rng.choice(ks, size=delta0.shape, p=pmf)).to(delta0.device)
    masks = torch.stack([masks & 2 ** i != 0 for i in range(len(deltas))])

    final_delta = torch.zeros_like(delta0)
    for mask, delta in zip(masks, deltas):
        final_delta[mask] += delta[mask]
    return final_delta / masks.sum(0).clamp(1) / (1 - probability)


def overlapping_sets_pmf(n, p, overlap, overlap_emphasis):
    if np.isclose(overlap, int(overlap)):
        if overlap % 2 == 0:
            pmf = np.array([1/n*float(bin(i).count("1") == 1) for i in range(1, 2**n)])
        else:
            pmf = np.array([0 for _ in range(1, 2**n - 1)] + [1])
    else:
        if math.floor(overlap) % 2 == 1:
            overlap = -overlap

        tan_overlap = np.tan(np.pi * (overlap - 0.5))
        pmf = np.zeros(2 ** n - 1)
        for i in range(1, 2 ** n):
            num_sets = bin(i).count("1")
            pmf[i-1] = tan_overlap*(num_sets - n/2)
        pmf = np.exp(pmf) / np.sum(np.exp(pmf))

    binomial_pmf = binom.pmf(np.arange(1, n + 1), n, p)
    expanded_binomial_pmf = np.zeros(2 ** n - 1)
    for i in range(1, 2 ** n):
        num_sets = bin(i).count("1")
        expanded_binomial_pmf[i-1] = binomial_pmf[num_sets-1] / binomial_coefficient_np(n, num_sets)
    expanded_binomial_pmf /= expanded_binomial_pmf.sum()

    pmf = weighted_sum.__wrapped__(
        pmf,
        weighted_sum.__wrapped__(pmf, expanded_binomial_pmf, alpha=1-abs(2*overlap-1)),
        alpha=overlap_emphasis,
    )
    return np.concatenate([[p], pmf * (1 - p)])


def binomial_coefficient_np(n, k):
    if k > n - k:
        k = n - k
    result = np.int64(1)
    for i in range(1, k+1):
        result = result * (n - i + 1) // i
    return result



import functools
import fuzzywuzzy.process
import inspect
import pathlib
import textwrap
import torch
from sd_mecha.recipe_nodes import MergeSpace, RecipeNode, ModelRecipeNode, MergeRecipeNode
from sd_mecha.hypers import Hyper
from typing import Optional, Callable, Dict, Tuple, TypeVar, Generic, get_type_hints, get_origin, Union, get_args, List, Set, Iterable


RecipeNodeOrPath = RecipeNode | str | pathlib.Path
T = TypeVar("T")


class LiftFlag(Generic[T]):
    def __init__(self):
        raise TypeError


class MergeMethod:
    create_recipe: Callable

    def __init__(self, f: Callable, volatile_hypers: Iterable[str]):
        self.__f = f
        self.__name = f.__name__
        self.__volatile_hypers = volatile_hypers
        self.__validate_f()

    def __validate_f(self):
        spec = inspect.getfullargspec(self.__f)
        if spec.defaults:
            params_with_default = spec.args[-len(spec.defaults):]
            raise TypeError(f"Default arguments are not supported for positional parameters. To declare hyperparameters, they must be keyword-only ({params_with_default})")

        for volatile_hyper in self.__volatile_hypers:
            if volatile_hyper not in spec.kwonlydefaults:
                raise TypeError(f"Keyword-only parameter '{volatile_hyper}' was marked as volatile but it is missing or does not have a default value")

        if spec.varkw is None:
            raise TypeError(f"**kwargs must be included in the function parameters")

    def __call__(self, inputs: Tuple[torch.Tensor, ...], hypers: Dict[str, Hyper], key: str, device: str, dtype: torch.dtype):
        args, kwargs = self.__get_args_kwargs(inputs, hypers, key, device, dtype)
        return self.__f(*args, **kwargs)

    def __get_args_kwargs(
        self,
        inputs: Tuple[torch.Tensor, ...],
        hypers: Dict[str, float],
        key: str,
        device: str,
        dtype: Optional[torch.dtype],
    ) -> Tuple[Tuple[torch.Tensor, ...], Dict]:
        if dtype is None:
            to_args = device,
        else:
            to_args = device, dtype

        for k in hypers:
            if k not in self.get_hyper_names():
                raise ValueError(f"method {self.__name} does not have a hyperparameter '{k}'")

        merge_method_args = tuple(
            v.to(*to_args)
            for v in inputs
        )
        return merge_method_args, hypers | {
            "device": device,
            "dtype": dtype,
            "key": key,
        }

    def get_return_merge_space(self, merge_spaces_args: List[MergeSpace]) -> MergeSpace:
        type_hints = get_type_hints(self.__f)
        model_names = self.get_model_names()
        varargs_name = self.get_model_varargs_name()
        if varargs_name is not None:
            model_names.extend([varargs_name]*(len(merge_spaces_args) - len(model_names)))

        resolved_input_spaces = {}
        for param, merge_space_arg in zip(model_names, merge_spaces_args):
            annotation = type_hints.get(param)
            if annotation:
                merge_space_param, key = self.__extract_lift_flag(annotation, param)

                if key in resolved_input_spaces:
                    # occurrence of already seen type var
                    if merge_space_arg != resolved_input_spaces[key]:
                        raise TypeError(f"parameter '{param}' of method {self.__name} expects {resolved_input_spaces[key]} but got {merge_space_arg}")
                elif merge_space_arg & merge_space_param:
                    resolved_input_spaces[key] = merge_space_arg
                else:
                    raise TypeError(f"parameter '{param}' of method {self.__name} expects {merge_space_param} but got {merge_space_arg}")

        merge_space_param, key = self.__extract_lift_flag(type_hints.get("return"), None)
        if key in resolved_input_spaces:
            return resolved_input_spaces[key]
        else:
            return merge_space_param

    def get_input_merge_spaces(self) -> Tuple[List[MergeSpace], Optional[MergeSpace]]:
        type_hints = get_type_hints(self.__f)
        model_names = self.get_model_names()
        merge_spaces = []
        for param in model_names:
            annotation = type_hints.get(param)
            if annotation:
                merge_space, _ = self.__extract_lift_flag(annotation, param)
                merge_spaces.append(merge_space)

        varargs_name = self.get_model_varargs_name()
        varargs_merge_space = None
        if varargs_name:
            annotation = type_hints.get(varargs_name)
            if annotation:
                varargs_merge_space, _ = self.__extract_lift_flag(annotation, varargs_name)

        return merge_spaces, varargs_merge_space

    def __extract_lift_flag(self, annotation, param) -> Tuple[MergeSpace, object]:
        if get_origin(annotation) is Union:
            for arg in get_args(annotation):
                if get_origin(arg) is LiftFlag:
                    return get_args(arg)[0], param
                elif isinstance(arg, TypeVar):
                    return get_args(arg.__bound__)[0], arg

    def get_model_names(self) -> List[str]:
        return inspect.getfullargspec(self.__f).args

    def get_hyper_names(self) -> Set[str]:
        return set(inspect.getfullargspec(self.__f).kwonlyargs)

    def get_volatile_hyper_names(self) -> Set[str]:
        return set(self.__volatile_hypers)

    def get_default_hypers(self) -> Dict[str, Hyper]:
        return inspect.getfullargspec(self.__f).kwonlydefaults or {}

    def get_model_varargs_name(self) -> Optional[str]:
        return inspect.getfullargspec(self.__f).varargs

    def get_name(self) -> str:
        return self.__name


def convert_to_recipe(
    f: Optional[Callable] = None, *,
    volatile_hypers: Iterable[str] = (),
    register: bool = True,
):
    if f is None:
        return lambda f: __convert_to_recipe_impl(f, volatile_hypers=volatile_hypers, register=register)
    return __convert_to_recipe_impl(f, volatile_hypers=volatile_hypers, register=register)


def __convert_to_recipe_impl(
    f: Callable, *,
    volatile_hypers: Iterable[str],
    register: bool,
):
    merge_method = MergeMethod(f, volatile_hypers)
    default_hypers = merge_method.get_default_hypers()

    model_params = "".join(
        f"{model_name}: {MergeRecipeNode.__name__}, "
        for model_name in merge_method.get_model_names()
    )
    hyper_params = "".join(
        f"{hyper_name}: {Hyper.__name__} = default_hypers[\"{hyper_name}\"], "
        if hyper_name in default_hypers else
        f"{hyper_name}: {Hyper.__name__}, "
        for hyper_name in merge_method.get_hyper_names()
    )

    model_args = "".join(
        f"{path_to_node.__name__}({model_name}), "
        for model_name in merge_method.get_model_names()
    )
    hyper_args = "".join(
        f"{hyper_name}={hyper_name}, "
        for hyper_name in merge_method.get_hyper_names() - merge_method.get_volatile_hyper_names()
    )
    volatile_hyper_args = "".join(
        f"{hyper_name}={hyper_name}, "
        for hyper_name in merge_method.get_volatile_hyper_names()
    )

    fn_locals = {}
    fn_globals = globals() | {
        "merge_method": merge_method,
        "default_hypers": default_hypers,
        "dtype": torch.dtype,  # `torch.dtype.__name__`
    }
    exec(textwrap.dedent(f"""
        def {f.__name__}(
            {model_params}
            *args: {MergeRecipeNode.__name__},
            {hyper_params}
            device: {Optional.__name__}[{str.__name__}] = None,
            dtype: {Optional.__name__}[{torch.dtype.__name__}] = None,
            **kwargs,
        ):
            return {MergeRecipeNode.__name__}(
                merge_method,
                {model_args}
                *({path_to_node.__name__}(arg) for arg in args),
                hypers=dict({hyper_args}),
                volatile_hypers=dict({volatile_hyper_args}),
                device=device,
                dtype=dtype,
            )
    """), fn_globals, fn_locals)
    res = fn_locals[f.__name__]
    res.__wrapped__ = f
    res.__wrapped_method__ = merge_method
    merge_method.create_recipe = res
    if register:
        _merge_methods_registry[f.__name__] = merge_method
    return res


_merge_methods_registry = {}


def resolve(identifier: str) -> MergeMethod:
    try:
        return _merge_methods_registry[identifier]
    except KeyError as e:
        suggestion = fuzzywuzzy.process.extractOne(str(e), _merge_methods_registry.keys())[0]
        raise ValueError(f"unknown merge method: {e}. Nearest match is '{suggestion}'")


@functools.cache
def path_to_node(node_or_path: RecipeNodeOrPath) -> RecipeNode:
    if isinstance(node_or_path, (str, pathlib.Path)):
        return ModelRecipeNode(node_or_path)
    return node_or_path


def clear_model_paths_cache():
    path_to_node.cache_clear()



import pathlib
import fuzzywuzzy.process
from typing import List, Optional
from sd_mecha import extensions, recipe_nodes
from sd_mecha.recipe_nodes import RecipeNode, ModelRecipeNode, MergeSpace, ParameterRecipeNode, RecipeVisitor


def deserialize_path(recipe: str | pathlib.Path, models_dir: Optional[str | pathlib.Path] = None) -> RecipeNode:
    if not isinstance(recipe, pathlib.Path):
        recipe = pathlib.Path(recipe)
    if recipe.exists():
        if recipe.suffix == ".safetensors":
            return recipe_nodes.ModelRecipeNode(recipe)
        elif recipe.suffix == ".mecha":
            return deserialize(recipe)
        else:
            raise ValueError(f"unable to deserialize '{recipe}': unknown extension")

    if models_dir is not None and not recipe.is_absolute():
        recipe = models_dir / recipe
    if not recipe.suffix:
        recipe = recipe.with_suffix(".safetensors")
    elif recipe.suffix != ".safetensors":
        raise ValueError(f"invalid path to safetensors or recipe: {recipe}")
    return recipe_nodes.ModelRecipeNode(recipe)


def deserialize(recipe: List[str] | str | pathlib.Path) -> RecipeNode:
    if not isinstance(recipe, list):
        with open(recipe, "r") as recipe:
            recipe = recipe.readlines()

    results = []

    def parse(line):
        command, *args = tokenize(line.strip())
        positional_args, named_args = preprocess_args(args)
        if command == "dict":
            results.append(dict(*positional_args, **named_args))
        elif command == "model":
            results.append(ModelRecipeNode(*positional_args, **named_args))
        elif command == "parameter":
            merge_space = MergeSpace[positional_args[1]]
            results.append(ParameterRecipeNode(positional_args[0], merge_space, **named_args))
        elif command == "merge":
            method, *positional_args = positional_args
            method = extensions.merge_method.resolve(method)
            results.append(method.create_recipe(*positional_args, **named_args))
        else:
            raise ValueError(f"unknown command: {command}")

    def preprocess_args(args):
        positional_args = []
        named_args = {}
        for arg_index, arg in enumerate(args):
            if '=' in arg:
                # note: this is wrong if "=" is inside quotes
                # however, quoted kwarg values (aka string hypers) will raise an exception in the constructor of recipes
                # so I'm not gonna bother fixing this with the tokenizer until it is actually useful to do so
                key, value = arg.split('=')
                named_args[key] = get_arg_value(value, arg_index)
            else:
                positional_args.append(get_arg_value(arg, arg_index))
        return positional_args, named_args

    def get_arg_value(arg, arg_index):
        try:
            if arg.startswith('&'):
                ref_index = int(arg[1:])
                if ref_index < 0 or ref_index >= len(results):
                    raise ValueError(f"reference {arg} out of bounds")
                return results[ref_index]
            elif arg.startswith('"') and arg.endswith('"'):
                return arg[1:-1]
            elif '.' in arg or 'e' in arg.lower():
                return float(arg)
            else:
                return int(arg)
        except ValueError as e:
            raise ValueError(f"argument {arg_index}: {str(e)}")

    def tokenize(line):
        tokens = []
        current_token = []
        inside_quotes = False
        for char in line:
            if char == '"':
                inside_quotes = not inside_quotes
                if not inside_quotes:  # End of quoted string
                    tokens.append(f'"{"".join(current_token)}"')
                    current_token = []
                continue
            elif char == ' ' and not inside_quotes:
                if current_token:  # End of a token
                    tokens.append(''.join(current_token))
                    current_token = []
                continue
            current_token.append(char)
        if inside_quotes:  # Handle mismatched quotes
            raise ValueError(f"mismatched quotes in input")
        if current_token:  # Add last token if exists
            tokens.append(''.join(current_token))
        return tokens

    for line_num, line in enumerate(recipe, 1):
        try:
            parse(line)
        except ValueError as e:
            raise ValueError(f"line {line_num}: {e}.\n    {line}")

    return results[-1]


def serialize(recipe: RecipeNode) -> str:
    serializer = SerializerVisitor()
    recipe.accept(serializer)
    return "\n".join(serializer.instructions)


class SerializerVisitor(RecipeVisitor):
    def __init__(self, instructions: Optional[List[str]] = None):
        self.instructions = instructions if instructions is not None else []

    def visit_model(self, node: recipe_nodes.ModelRecipeNode) -> int:
        line = f'model "{node.path}" "{node.model_arch.identifier}" "{node.model_type.identifier}"'
        return self.__add_instruction(line)

    def visit_parameter(self, node: recipe_nodes.ParameterRecipeNode) -> int:
        merge_space = str(MergeSpace.BASE).split(".")[1]
        model_arch = f' "{node.model_arch.identifier}"' if node.model_arch else ""
        line = f'parameter "{node.name}" "{merge_space}"' + model_arch
        return self.__add_instruction(line)

    def visit_merge(self, node: recipe_nodes.MergeRecipeNode) -> int:
        models = [
            f" &{model.accept(self)}"
            for model in node.models
        ]

        hypers = []
        for hyper_k, hyper_v in node.hypers.items():
            if isinstance(hyper_v, dict):
                dict_line = "dict" + "".join(f" {k}={v}" for k, v in hyper_v.items())
                hyper_v = f"&{self.__add_instruction(dict_line)}"
            hypers.append(f" {hyper_k}={hyper_v}")

        line = f'merge "{node.merge_method.get_name()}"{"".join(models)}{"".join(hypers)}'
        return self.__add_instruction(line)

    def __add_instruction(self, instruction: str) -> int:
        try:
            return self.instructions.index(instruction)
        except ValueError:
            self.instructions.append(instruction)
            return len(self.instructions) - 1



import dataclasses
import functools
import operator

import fuzzywuzzy.process
import pathlib
import re
import traceback
import yaml
from typing import Set, Dict, List

WILDCARD = "*"
PADDING = "#"


def discover_block_prefixes(keys: Set[str], config: dict, version_id: str):
    discovered_blocks = {}

    for module, module_config in config["merge"].items():
        prefix = module_config["prefix"]
        for shorthand, patterns in module_config["blocks"].items():
            if isinstance(patterns, str):
                patterns = [patterns]

            first_pattern_re = re.escape(patterns[0]).replace(re.escape(WILDCARD), r'(\w+)').replace(re.escape(PADDING), r'\w+')
            pattern = re.compile(rf"^{prefix}\.{first_pattern_re}")

            for key in keys:
                match = pattern.match(key)
                if match:
                    block_id = match.groups()[0] if WILDCARD in patterns[0] else ""
                    block_shorthand = (version_id + "_" + module + "_block_" + shorthand.replace(WILDCARD, "{}")).format(block_id)
                    if block_shorthand not in discovered_blocks:
                        discovered_blocks[block_shorthand] = {
                            "patterns": [re.compile(p) for p in sorted((
                                (re.escape(f"{prefix}.{p.replace(WILDCARD, block_id)}") + r"(?:\.|$)").replace(re.escape(PADDING), r'\w+')
                                for p in patterns
                            ), key=lambda s: len(s.split(".")), reverse=True)],
                            "module": module,
                        }

    return discovered_blocks


def discover_blocks(keys, discovered_block_prefixes, version_id):
    blocks = {}
    classes = {}
    for key in keys:
        for block, prefixes in discovered_block_prefixes.items():
            if any((match := prefix.match(key)) for prefix in prefixes["patterns"]):
                blocks.setdefault(key, set()).add(block)
                clazz = key[len(match.group(0)):]
                while not clazz or clazz in ["weight", "bias"] or clazz[:1].isnumeric():
                    if clazz:
                        part = key[:-len(clazz)-1].split(".")[-1]
                    else:
                        part = match.group(0).split(".")[-1]
                    clazz = part + "." + clazz
                clazz = clazz.replace(".weight", "").replace(".bias", "").replace(".", "_")
                classes[key] = {version_id + "_" + prefixes["module"] + "_class_" + clazz}
                break

    return blocks, classes


@dataclasses.dataclass
class ModelArch:
    identifier: str
    components: Set[str]
    keys: Set[str]
    keys_to_forward: Set[str]
    keys_to_merge: Set[str]
    blocks: Dict[str, Set[str]]
    classes: Dict[str, Set[str]]
    location: str

    def user_keys(self) -> Set[str]:
        return (
            functools.reduce(operator.or_, self.blocks.values(), set()) |
            functools.reduce(operator.or_, self.classes.values(), set())
        )


def register_model_arch(
    yaml_config_path: str | pathlib.Path,
    identifier: str,
):
    stack_frame = traceback.extract_stack(None, 2)[0]

    if identifier in _model_archs_registry:
        existing_location = _model_archs_registry[identifier].location
        raise ValueError(f"Extension '{identifier}' is already registered at {existing_location}.")

    if isinstance(yaml_config_path, str):
        yaml_config_path = pathlib.Path(yaml_config_path)

    with open(yaml_config_path, "r") as f:
        config = yaml.safe_load(f.read())

    keys = config["keys"]
    if isinstance(keys, str):
        with open(yaml_config_path.parent / keys, "r") as f:
            keys = [k.strip() for k in f.readlines()]

    prefixes_to_forward = tuple(config["passthrough"])
    keys_to_forward = set(key for key in keys if key.startswith(prefixes_to_forward))
    keys_to_merge = set(key for key in keys if key not in keys_to_forward)
    block_prefixes = discover_block_prefixes(keys_to_merge, config, identifier)
    blocks, classes = discover_blocks(keys_to_merge, block_prefixes, identifier)
    components = set(config["merge"].keys())

    location = f"{stack_frame.filename}:{stack_frame.lineno}"
    _model_archs_registry[identifier] = ModelArch(identifier, components, keys, keys_to_forward, keys_to_merge, blocks, classes, location)


_model_archs_registry = {}


def resolve(identifier: str) -> ModelArch:
    try:
        return _model_archs_registry[identifier]
    except KeyError as e:
        suggestion = fuzzywuzzy.process.extractOne(str(e), _model_archs_registry.keys())[0]
        raise ValueError(f"unknown model architecture: {e}. Nearest match is '{suggestion}'")


def get_all() -> List[str]:
    return list(_model_archs_registry.keys())



import pathlib
from sd_mecha.extensions.model_arch import register_model_arch

models_dir = pathlib.Path(__file__).parent / "models"
register_model_arch(models_dir / "sd1_ldm.yaml", "sd1")
register_model_arch(models_dir / "sdxl_sgm.yaml", "sdxl")
